@@ -156,19 +156,24 @@ def add_images_to_dictionary(images_dictionary, sitk_dicom_reader, path: typing.
             slice_thickness = float(sitk_dicom_reader.GetMetaData("0018|0050"))
         study_instance_uid = sitk_dicom_reader.GetMetaData("0020|000d")
         temp_dict = {'PatientID': patientID, 'SeriesInstanceUID': series_instance_uid,
-                     'StudyInstanceUID': study_instance_uid, 'RTs': {}, 'RDs': {},
+                     'StudyInstanceUID': study_instance_uid, 'RTs': {}, 'RDs': {}, 'RPs': {},
                      'Image_Path': path, 'Description': description, 'Pixel_Spacing_X': pixel_spacing_x,
                      'Pixel_Spacing_Y': pixel_spacing_y, 'Slice_Thickness': slice_thickness}
         images_dictionary[series_instance_uid] = temp_dict
 
 
 def add_rp_to_dictionary(ds, path: typing.Union[str, bytes, os.PathLike], rp_dictionary):
-    series_instance_uid = ds.SeriesInstanceUID
-    if series_instance_uid not in rp_dictionary:
-        for referenced_structureset in ds.ReferencedStructureSetSequence:
-            refed_structure_uid = referenced_structureset.ReferencedSOPInstanceUID
-            temp_dict = {'Path': path, 'ReferencedStructureUID': refed_structure_uid}
+    try:
+        series_instance_uid = ds.SeriesInstanceUID
+        if series_instance_uid not in rp_dictionary:
+            refed_structure_uid = ds.ReferencedStructureSetSequence[0].ReferencedSOPInstanceUID
+            refed_dose_uid = ds.DoseReferenceSequence[0].DoseReferenceUID
+            temp_dict = {'Path': path, 'SOPInstanceUID': ds.SOPInstanceUID,
+                         'ReferencedStructureSetSOPInstanceUID': refed_structure_uid,
+                         'ReferencedDoseSOPUID': refed_dose_uid, 'Description': ds.StudyDescription}
             rp_dictionary[series_instance_uid] = temp_dict
+    except:
+        print("Had an error loading " + path)
 
 
 def add_rt_to_dictionary(ds, path: typing.Union[str, bytes, os.PathLike], rt_dictionary):
@@ -176,25 +181,30 @@ def add_rt_to_dictionary(ds, path: typing.Union[str, bytes, os.PathLike], rt_dic
     :param ds: pydicom data structure
     :param path: path to the images or structure in question
     """
-    series_instance_uid = ds.SeriesInstanceUID
-    if series_instance_uid not in rt_dictionary:
-        for referenced_frame_of_reference in ds.ReferencedFrameOfReferenceSequence:
-            for referred_study_sequence in referenced_frame_of_reference.RTReferencedStudySequence:
-                for referred_series in referred_study_sequence.RTReferencedSeriesSequence:
-                    refed_series_instance_uid = referred_series.SeriesInstanceUID
-                    if Tag((0x3006, 0x020)) in ds.keys():
-                        ROI_Structure = ds.StructureSetROISequence
-                    else:
-                        ROI_Structure = []
-                    rois_in_structure = {}
-                    rois = []
-                    for Structures in ROI_Structure:
-                        rois.append(Structures.ROIName.lower())
-                        if Structures.ROIName not in rois_in_structure:
-                            rois_in_structure[Structures.ROIName] = Structures.ROINumber
-                    temp_dict = {'Path': path, 'ROI_Names': rois, 'ROIs_in_structure': rois_in_structure,
-                                 'SeriesInstanceUID': refed_series_instance_uid, 'Plans': []}
-                    rt_dictionary[series_instance_uid] = temp_dict
+    try:
+        series_instance_uid = ds.SeriesInstanceUID
+        sop_instance_uid = ds.SOPInstanceUID
+        if series_instance_uid not in rt_dictionary:
+            for referenced_frame_of_reference in ds.ReferencedFrameOfReferenceSequence:
+                for referred_study_sequence in referenced_frame_of_reference.RTReferencedStudySequence:
+                    for referred_series in referred_study_sequence.RTReferencedSeriesSequence:
+                        refed_series_instance_uid = referred_series.SeriesInstanceUID
+                        if Tag((0x3006, 0x020)) in ds.keys():
+                            ROI_Structure = ds.StructureSetROISequence
+                        else:
+                            ROI_Structure = []
+                        rois_in_structure = {}
+                        rois = []
+                        for Structures in ROI_Structure:
+                            rois.append(Structures.ROIName.lower())
+                            if Structures.ROIName not in rois_in_structure:
+                                rois_in_structure[Structures.ROIName] = Structures.ROINumber
+                        temp_dict = {'Path': path, 'ROI_Names': rois, 'ROIs_in_structure': rois_in_structure,
+                                     'SeriesInstanceUID': refed_series_instance_uid, 'Plans': {}, 'Doses': {},
+                                     'SOPInstanceUID': sop_instance_uid}
+                        rt_dictionary[series_instance_uid] = temp_dict
+    except:
+        print("Had an error loading " + path)
 
 
 def add_rd_to_dictionary(sitk_dicom_reader, rd_dictionary):
@@ -202,15 +212,23 @@ def add_rd_to_dictionary(sitk_dicom_reader, rd_dictionary):
     :param ds: pydicom data structure
     :param path: path to the images or structure in question
     """
-    series_instance_uid = sitk_dicom_reader.GetMetaData("0020|000e")
-    if series_instance_uid not in rd_dictionary:
-        study_instance_uid = sitk_dicom_reader.GetMetaData("0020|000d")
-        description = None
-        if "0008|103e" in sitk_dicom_reader.GetMetaDataKeys():
-            description = sitk_dicom_reader.GetMetaData("0008|103e")
-        temp_dict = {'Path': sitk_dicom_reader.GetFileName(), 'StudyInstanceUID': study_instance_uid,
-                     'Description': description}
-        rd_dictionary[series_instance_uid] = temp_dict
+    try:
+        ds = pydicom.read_file(sitk_dicom_reader.GetFileName())
+        series_instance_uid = sitk_dicom_reader.GetMetaData("0020|000e")
+        rt_sopinstance_uid = ds.ReferencedStructureSetSequence[0].ReferencedSOPInstanceUID
+        rp_sopinstance_uid = ds.ReferencedRTPlanSequence[0].ReferencedSOPInstanceUID
+        if series_instance_uid not in rd_dictionary:
+            study_instance_uid = sitk_dicom_reader.GetMetaData("0020|000d")
+            description = None
+            if "0008|103e" in sitk_dicom_reader.GetMetaDataKeys():
+                description = sitk_dicom_reader.GetMetaData("0008|103e")
+            temp_dict = {'Path': sitk_dicom_reader.GetFileName(), 'StudyInstanceUID': study_instance_uid,
+                         'SOPInstanceUID': sitk_dicom_reader.GetMetaData("0008|0018"),
+                         'Description': description, 'ReferencedStructureSetSOPInstanceUID': rt_sopinstance_uid,
+                         'ReferencedPlanSOPInstanceUID': rp_sopinstance_uid}
+            rd_dictionary[series_instance_uid] = temp_dict
+    except:
+        print("Had an error loading " + sitk_dicom_reader.GetFileName())
 
 
 def add_sops_to_dictionary(sitk_dicom_reader, series_instances_dictionary):
@@ -280,6 +298,7 @@ class AddDicomToDictionary(object):
                 add_rt_to_dictionary(ds=rt, path=lstRSFile, rt_dictionary=rt_dictionary)
             elif modality.lower().find('plan') != -1:
                 add_rp_to_dictionary(ds=rt, path=lstRSFile, rp_dictionary=rp_dictionary)
+        xxx = 1
 
 
 class DicomReaderWriter(object):
@@ -384,23 +403,42 @@ class DicomReaderWriter(object):
                 template = return_template_dictionary()
                 template['RTs'].update({rt_series_instance_uid: self.rt_dictionary[rt_series_instance_uid]})
                 self.series_instances_dictionary[index] = template
-        study_instance_uids = None
         for rd_series_instance_uid in self.rd_dictionary:
-            if study_instance_uids is None:
-                study_instance_uids = []
-                for key, value in self.series_instances_dictionary.items():
-                    study_instance_uids.append(value['StudyInstanceUID'])
-            study_instance_uid = self.rd_dictionary[rd_series_instance_uid]['StudyInstanceUID']
-            if study_instance_uid in study_instance_uids:
-                index = study_instance_uids.index(study_instance_uid)
-                self.series_instances_dictionary[index]['RDs'].update({rd_series_instance_uid:
-                                                                           self.rd_dictionary[rd_series_instance_uid]})
-            else:
+            added = False
+            struct_ref = self.rd_dictionary[rd_series_instance_uid]['ReferencedStructureSetSOPInstanceUID']
+            for image_series_key in self.series_instances_dictionary:
+                rts = self.series_instances_dictionary[image_series_key]['RTs']
+                for rt_key in rts:
+                    structure_sop_uid = rts[rt_key]['SOPInstanceUID']
+                    if struct_ref == structure_sop_uid:
+                        rts[rt_key]['Doses'][rd_series_instance_uid] = self.rd_dictionary[rd_series_instance_uid]
+                        self.series_instances_dictionary[image_series_key]['RDs'].update({rd_series_instance_uid:
+                                                                                              self.rd_dictionary[rd_series_instance_uid]})
+                    added = True
+            if not added:
                 while index in self.series_instances_dictionary:
                     index += 1
                 template = return_template_dictionary()
                 template['RDs'].update({rd_series_instance_uid: self.rd_dictionary[rd_series_instance_uid]})
                 self.series_instances_dictionary[index] = template
+        for rp_series_instance_uid in self.rp_dictionary:
+            added = False
+            struct_ref = self.rp_dictionary[rp_series_instance_uid]['ReferencedStructureSetSOPInstanceUID']
+            for image_series_key in self.series_instances_dictionary:
+                rts = self.series_instances_dictionary[image_series_key]['RTs']
+                for rt_key in rts:
+                    structure_sop_uid = rts[rt_key]['SOPInstanceUID']
+                    if struct_ref == structure_sop_uid:
+                        rts[rt_key]['Plans'][rp_series_instance_uid] = self.rp_dictionary[rp_series_instance_uid]
+                        self.series_instances_dictionary[image_series_key]['RPs'].update({rp_series_instance_uid:
+                                                                                              self.rp_dictionary[rp_series_instance_uid]})
+                    added = True
+            if not added:
+                while index in self.series_instances_dictionary:
+                    index += 1
+                template = return_template_dictionary()
+                template['RPs'].update({rp_series_instance_uid: self.rp_dictionary[rp_series_instance_uid]})
+                self.series_instances_dictionary[index] = template
 
     def __reset__(self):
         self.__reset_RTs__()
@@ -525,6 +563,28 @@ class DicomReaderWriter(object):
                 out_file_paths += [image_dictionary[key][structure_key]['Path']]
         return out_file_paths
 
+    def return_files_from_index(self, index: int) -> List[str]:
+        """
+        Args:
+            index: An integer index found in images_dictionary.
+
+        Returns:
+            file_list: A list of file paths that are associated with that index, being images, RTs, RDs, and RPs
+        """
+        out_file_paths = list()
+        image_dictionary = self.series_instances_dictionary[index]
+        UID = image_dictionary['SeriesInstanceUID']
+        dicom_path = image_dictionary['Image_Path']
+        image_reader = sitk.ImageFileReader()
+        image_reader.LoadPrivateTagsOn()
+        reader = sitk.ImageSeriesReader()
+        reader.GlobalWarningDisplayOff()
+        out_file_paths += reader.GetGDCMSeriesFileNames(dicom_path, UID)
+        for key in ['RTs', 'RDs', 'RPs']:
+            for structure_key in image_dictionary[key]:
+                out_file_paths += [image_dictionary[key][structure_key]['Path']]
+        return out_file_paths
+
     def where_are_RTs(self, ROIName: str) -> None:
         print('Please move over to using .where_is_ROI(), as this better represents the definition')
         self.where_is_ROI(ROIName=ROIName)
